unit HLDateCvt;

{$B-}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db;

type
  TCheckDateEvent = procedure
    (Field: TField; const Text: string; var IsValid: Boolean) of object;

  THLDateConverter = class(TComponent)
  private
    FAllowEmptyDate: Boolean;
    FDisplayFormat: string;
    FEditMask: string;
    FDataSets: TList;
    FFields: TList;     // Date fields
    FOnCheckDate: TCheckDateEvent;

    procedure FreeFieldListItems;
    procedure DefaultCheckDate(AField: TField; DateStr: string;
      var IsValid: Boolean);

    procedure SetDisplayFormat(Value: string);
  protected
    procedure Loaded; override;
    procedure HLGetText(Sender: TField; var Text: string; DisplayText: Boolean); dynamic;
    procedure HLSetText(Sender: TField; const Text: string); dynamic;
    procedure GetRocDate(Sender: TField; var Text: string; DisplayText: Boolean);
    procedure SetRocDate(Sender: TField; const Text: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure Add(DataSet: TDataSet);
    procedure AddAll(DSOwner: TComponent);
    procedure Convert;
  published
    property AllowEmptyDate: Boolean read FAllowEmptyDate write FAllowEmptyDate default true;
    property DisplayFormat: string read FDisplayFormat write SetDisplayFormat;
    property OnCheckDate: TCheckDateEvent read FOnCheckDate write FOnCheckDate;
  end;

var
  g_RocYearDigits: integer;     // إ~
  g_DateFormat: string;      // إ榡


procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Flyup', [THLDateConverter]);
end;

function StrPadL(const S: string; Len: Integer; ch: char): string;
var
  iSrcLength: byte;
  i: integer;
  tmp: string;
begin
  iSrcLength := Length(S);
  if iSrcLength > Len then
  begin
    Result := Copy(S, 1, Len);
    Exit;
  end;
  Result := TrimLeft(S);
  iSrcLength := Length(Result);
  tmp := '';
  for i := 1 to Len-iSrcLength do
    tmp := tmp + ch;
  Result := tmp + Result;
end;

function CheckRocDateStr(DateStr: string): boolean;
var
  iYear, iMonth, iDay: word;
  DayTable: PDayTable;
  i: Integer;
begin
  Result := False;

  // 榡ˬd
  if Length(DateStr) <> (g_RocYearDigits+6) then
    Exit;
  // NϥΪ̥iϥΨ쪺jrΤ@ഫ DateSeparator
  for i := 1 to Length(DateStr) do
  begin
    if (DateStr[i] = '.') or (DateStr[i] = '-') then
      DateStr[i] := DateSeparator;
  end;
  if (DateStr[g_RocYearDigits+1] <> DateSeparator) or
     (DateStr[g_RocYearDigits+4] <> DateSeparator) then
    Exit;

  //  -> 褸
  try
    iYear := StrToInt(Trim(Copy(DateStr, 1, g_RocYearDigits))) + 1911;
    iMonth := StrToInt(Trim(Copy(DateStr, g_RocYearDigits+2, 2)));
    iDay := StrToInt(Trim(Copy(DateStr, g_RocYearDigits+5, 2)));
  except
    Exit;
  end;

  DayTable := @MonthDays[IsLeapYear(iYear)]; // MonthDays wq SysUtils.pas
  if (iYear >= 1) and (iYear <= 9999) and (iMonth >= 1) and
     (iMonth <= 12) and (iDay >= 1) and (iDay <= DayTable^[iMonth]) then
    Result := true;
end;


constructor THLDateConverter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);

  FOnCheckDate := nil;
  FDataSets := TList.Create;
  FFields := TList.Create;

  FAllowEmptyDate := true;
  FDisplayFormat := '';

  FEditMask := '!' + Copy('9999', 1, g_RocYearDigits) + DateSeparator +
    '99' + DateSeparator + '99;1;_';
end;

destructor THLDateConverter.Destroy;
begin
  if FFields <> nil then
  begin
    FreeFieldListItems;
    FFields.Free;
    FFields := nil;
  end;

  if FDataSets <> nil then
  begin
    FDataSets.Free;
    FDataSets := nil;
  end;

  inherited Destroy;
end;

procedure THLDateConverter.Loaded;
begin
  inherited Loaded;
  { ]ഫ榡|dIƶƥAG]pɴഫʧ@ }
  if csDesigning in ComponentState then
    Exit;
  AddAll(Owner);
  Convert;
end;

procedure THLDateConverter.DefaultCheckDate(AField: TField; DateStr: string;
  var IsValid: Boolean);
var
  s: string;

  function IsEmptyDateStr(sDate: string): Boolean;
  var
    i: integer;
  begin
    Result := true;
    for i := 1 to Length(sDate) do
    begin
      if not (sDate[i] in [' ', '-', '/', '.']) then
      begin
        Result := false;
        break;
      end;
    end;
  end;
begin
  IsValid := true;
  if IsEmptyDateStr(DateStr) then
  begin
    if (FAllowEmptyDate) then
      Exit;
    s := AField.DisplayLabel + 'ରť!';
    Application.MessageBox(PChar(s), 'T', MB_OK or MB_ICONINFORMATION);
    IsValid := false;
    Exit;
  end;
  if not CheckRocDateStr(DateStr) then
  begin
    s := QuotedStr(DateStr) + 'OXk!';
    Application.MessageBox(PChar(s), '~', MB_OK or MB_ICONERROR);
    IsValid := false;
  end;
end;

procedure THLDateConverter.SetDisplayFormat(Value: string);
var
  i: integer;
  pField: TField;
begin
  if FDisplayFormat = Value then
    Exit;
  FDisplayFormat := Value;

  if FFields = nil then
    Exit;
  // sDisplayFormats]wC쪫
  for i := 0 to (FFields.Count - 1) do
  begin
    pField := FFields.Items[i];
    if (pField = nil) then
      continue;
    if (pField is TDateTimeField) then
      TDateTimeField(pField).DisplayFormat := FDisplayFormat;
  end;
end;

procedure THLDateConverter.Add(DataSet: TDataSet);
begin
  if FDataSets.IndexOf(DataSet) >= 0 then
    Exit;
  FDataSets.Add(DataSet);
end;

procedure THLDateConverter.AddAll(DSOwner: TComponent);
var
  i: integer;
begin
  for i := 0 to DSOwner.ComponentCount-1 do
  begin
    if DSOwner.Components[i] is TDataSet then
      Add(TDataSet(DSOwner.Components[i]));
  end;
end;

procedure THLDateConverter.Convert;
var
  i, j: integer;
  DataSet: TDataSet;
  pField: TField;
begin
  FreeFieldListItems;

  for i := 0 to FDataSets.Count-1 do
  begin
    DataSet := FDataSets.Items[i];
    for j := 0 to DataSet.FieldCount-1 do
    begin
      pField := DataSet.Fields[j];
      if (pField.DataType in [ftDate, ftDateTime]) then
      begin
        TDateTimeField(pField).DisplayFormat := FDisplayFormat;
        pField.EditMask := FEditMask;
        pField.OnGetText := HLGetText;
        pField.OnSetText := HLSetText;

        FFields.Add(pField);
      end;
    end;
  end;
end;

procedure THLDateConverter.FreeFieldListItems;
var
  i: integer;
  pField: TField;
begin
  if FFields = nil then
    Exit;
  for i := 0 to (FFields.Count - 1) do
  begin
    pField := FFields.Items[i];
    if (pField = nil) then
      continue;
    try
      if (pField.DataType in [ftDate, ftDateTime]) then
      begin
        TDateTimeField(pField).DisplayFormat := '';
        pField.EditMask := '';
        pField.OnGetText := nil;
        pField.OnSetText := nil;
      end;
    except
      // ignore;
    end;
  end;
  FFields.Clear;
end;

(*--------------------------------------------------------------------------
  GetRocDate

    禡Ω TField.OnGetText ƥ, γ~ON쪺ഫإ
    榡, åH Text Ǧ^CpGഫL{oͿ~, hǦ^ťդrC

  Ѽ
  
    - Text    榡 ~// (: 086/08/17)
--------------------------------------------------------------------------*)
procedure THLDateConverter.GetRocDate(Sender: TField; var Text: string;
  DisplayText: Boolean);
var
  TheDate: TDateTime;
  wYear, wMonth, wDay: WORD;
  F: string;
begin
  try
    TheDate := Sender.AsDateTime;
    DecodeDate(TheDate, wYear, wMonth, wDay);
    if (wYear < 1911) or (wMonth < 1) or (wDay < 1) then
      Abort;
    if (DisplayText) and (FDisplayFormat <> '') then
      F := FDisplayFormat
    else
      F := g_DateFormat;
    DateTimeToString(Text, F, TheDate);
    if (F[1] = 'e') then
      Text := StrPadL(Text, Length(F), '0');
  except
    Text := Copy('    ', 1, g_RocYearDigits) + DateSeparator + '  ' +
      DateSeparator + '  ';
  end;
end;


(*--------------------------------------------------------------------------
  SetRocDate

    禡Ω TField.OnSetText ƥ, γ~ONѼ Text (إ)ഫ
    vHKsJƪC

  Ѽ

  - Text    榡 ~// (: 086/08/17)
--------------------------------------------------------------------------*)
procedure THLDateConverter.SetRocDate(Sender: TField; const Text: string);
var
  TheDate: TDateTime;
  wYear, wMonth, wDay: word;
  IsValid: Boolean;
begin
  IsValid := true;
  if Assigned(FOnCheckDate) then
    FOnCheckDate(Sender, Text, IsValid)
  else
    DefaultCheckDate(Sender, Text, IsValid);
  if not IsValid then
    Abort;

  try
    wYear := StrToInt(Trim(Copy(Text, 1, g_RocYearDigits))) + 1911;
    wMonth := StrToInt(Trim(Copy(Text, g_RocYearDigits+2, 2)));
    wDay := StrToInt(Trim(Copy(Text, g_RocYearDigits+5, 2)));
    TheDate := EncodeDate(wYear, wMonth, wDay);
    Sender.AsDateTime := TheDate;
  except
    Sender.AsString := '';
  end;
end;

procedure THLDateConverter.HLGetText(Sender: TField; var Text: string;
  DisplayText: Boolean);
begin
  GetRocDate(Sender, Text, DisplayText);
end;

procedure THLDateConverter.HLSetText(Sender: TField; const Text: string);
begin
  SetRocDate(Sender, Text);
end;

initialization
  g_RocYearDigits := 3;
  g_DateFormat := Copy('eeee', 1, g_RocYearDigits) + DateSeparator +
    'mm' + DateSeparator + 'dd';

end.
